/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.github.io/license/

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

#include "const.glsl"
#include "rho.glsl"

#define airScatterIterations 12
#define airmassIterations 6

struct airScatterData {
    vec3 rho;
    mat3 sunScattering;
    mat3 moonScattering;
};

airScatterData atmosInscatter(vec3 pos, vec3 dir, mat2x3 light, float dist, const uint steps, bool moonlight) {
    float rlength   = dist * rcp(float(steps));
    vec3 rstep      = dir * rlength;
        pos        += rstep * 0.25;

    vec3 transmittance = vec3(1.0);

    airScatterData total = airScatterData(vec3(0.0), mat3(0.0), mat3(0.0));

    for (uint i = 0; i < steps; ++i, pos += rstep) {
        vec3 rho    = airRho(planetCurveAltitude(pos));

        if (rho.y > 1e35) break;

        vec3 stepRho    = rho * rlength;
        vec3 od         = airExtinctMat * stepRho;
            rho        += stepRho;

        vec3 stept      = saturate(expf(-od));
        vec3 scatterInt = saturate((stept - 1.0) * rcp(-od));
        vec3 visScatter = transmittance * scatterInt;

        vec3 airmassSun = airmass(pos, light[0], airmassIterations);
        vec3 transmittanceSun = saturate(expf(-airExtinctMat * airmassSun));

        total.sunScattering[0] += visScatter * stepRho.x * transmittanceSun;
        total.sunScattering[1] += visScatter * stepRho.y * transmittanceSun;
        total.sunScattering[2] += visScatter * (airScatterMat * stepRho.xy) * transmittanceSun;

        if (moonlight) {
            vec3 airmassMoon = airmass(pos, light[1], airmassIterations);
            vec3 transmittanceMoon = saturate(expf(-airExtinctMat * airmassMoon));

            total.moonScattering[0] += visScatter * stepRho.x * transmittanceMoon;
            total.moonScattering[1] += visScatter * stepRho.y * transmittanceMoon;
            total.moonScattering[2] += visScatter * (airScatterMat * stepRho.xy) * transmittanceMoon;
        }

        transmittance  *= stept;
    }

    return total;
}

vec2 airPhaseFunction(float cosTheta) {
    return vec2(rayleighPhase(cosTheta), mieCS(cosTheta, airMieG));
}
float airMieBackscatter(float cosTheta) {
    return mieHG(cosTheta, -airMieG * rcp(pi)) * rcp(pi);
}

vec3 atmosphericScatter(vec3 dir, mat2x3 light, vec3 illuminance) {
    vec2 cosTheta = vec2(dot(dir, light[0]), dot(dir, light[1]));

    vec3 pos    = vec3(0.0, planetRad + eyeAltitude, 0.0);
    float dist  = airRSI(pos, dir);

    airScatterData scatterData = atmosInscatter(pos, dir, light, dist, airScatterIterations, daytime.w > 0.0);

    vec4 airPhase = vec4(airPhaseFunction(cosTheta.x), airPhaseFunction(cosTheta.y));

    vec3 sunScattering  = scatterData.sunScattering[0] * airRayleighCoeff * airPhase.x;
        sunScattering  += scatterData.sunScattering[1] * airMieCoeff * airPhase.y;
        sunScattering  += scatterData.sunScattering[2] * illuminance;
        sunScattering  *= sunIllum;
        sunScattering  *= mix(0.85, 0.75, sqrt(saturate(light[0].y * 2.0)));

    vec3 moonScattering = scatterData.moonScattering[0] * airRayleighCoeff * airPhase.x;
        moonScattering += scatterData.moonScattering[1] * airMieCoeff * airPhase.y;
        moonScattering += scatterData.moonScattering[2] * illuminance;
        moonScattering *= moonIllum * sqrt(daytime.w);

    return max(sunScattering + moonScattering, 0.0);
}