/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.github.io/license/

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

#include "/lib/util/poisson.glsl"

struct shadowData {
    float shadow;
    vec3 color;
    vec3 subsurfaceScatter;
};

float vectorAverage(vec4 x) {
    return dot(x, vec4(0.25));
}

const ivec2 shadowmapSize   = ivec2(shadowMapResolution * MC_SHADOW_QUALITY);
const vec2 shadowmapPixel   = 1.0 / vec2(shadowmapSize);

float shadowDefault(sampler2D tex, vec3 pos, vec2 offset) {
    return step(pos.z, textureLod(tex, pos.xy + offset, 0).x);
}

float shadowTexel(sampler2D tex, vec3 pos) {
    ivec2 location  = ivec2((pos.xy - (0.5 * shadowmapPixel)) * shadowmapSize);

    return step(pos.z, texelFetch(tex, location, 0).x);
}
float shadowTexel(sampler2D tex, vec3 pos, vec2 offset) {
    ivec2 location  = ivec2(((pos.xy + offset) - (0.5 * shadowmapPixel)) * shadowmapSize);

    return step(pos.z, texelFetch(tex, location, 0).x);
}
float shadowLinear(sampler2D tex, vec3 pos, vec2 offset) {
    //vec2 size       = textureSize(tex, 0);
    vec2 coord      = fract(pos.xy - (0.5 * shadowmapPixel) + offset);
    ivec2 location  = ivec2(coord * shadowmapSize);

    vec4 samples    = vec4(
        texelFetch(tex, location, 0).x,                texelFetch(tex, location + ivec2(1, 0), 0).x,
        texelFetch(tex, location +  ivec2(0, 1), 0).x, texelFetch(tex, location + ivec2(1, 1), 0).x
    );
        samples     = step(vec4(pos.z), samples);

    vec2 weights    = fract(coord * shadowmapSize);


    return mix(
        mix(samples.x, samples.y, weights.x),
        mix(samples.z, samples.w, weights.x), weights.y
    );
}

float textureShadowPCF(sampler2D tex, vec3 pos) {
    //vec2 size       = textureSize(tex, 0);
    vec2 coord      = fract(pos.xy - 0.5 * shadowmapPixel);
    ivec2 location  = ivec2(coord * shadowmapSize);

    vec4 samples    = vec4(
        texelFetch(tex, location, 0).x,                texelFetch(tex, location + ivec2(1, 0), 0).x,
        texelFetch(tex, location +  ivec2(0, 1), 0).x, texelFetch(tex, location + ivec2(1, 1), 0).x
    );
        samples     = step(vec4(pos.z), samples);

    vec2 weights    = fract(coord * shadowmapSize);


    return mix(
        mix(samples.x, samples.y, weights.x),
        mix(samples.z, samples.w, weights.x), weights.y
    );
}

float getShadowSoft(sampler2D tex, vec3 pos) {
    float step  = rcp(float(shadowMapResolution));
    float n     = ditherBluenoise()*pi;
    vec2 noise  = vec2(cos(n), sin(n));
    vec3 offset = vec3(noise, 0.0)*step;

    float s0    = textureShadowPCF(tex, pos)*0.5;
    float s1    = textureShadowPCF(tex, pos+offset);
    float s2    = textureShadowPCF(tex, pos-offset);

    return saturate((s0 + s1 + s2)*0.4);
}

float getShadowSoftHQ(sampler2D tex, vec3 pos) {
    float step  = rcp(float(shadowMapResolution));
    float n     = ditherBluenoise()*pi;
    vec2 noise  = vec2(cos(n), sin(n));
    vec3 offset = vec3(noise, 0.0)*step;

    float s0    = textureShadowPCF(tex, pos);
    float s1    = textureShadowPCF(tex, pos+offset);
    float s2    = textureShadowPCF(tex, pos-offset);
    float s3    = textureShadowPCF(tex, pos+offset*2.0)*0.5;
    float s4    = textureShadowPCF(tex, pos-offset*2.0)*0.5;

    return (s0 + s1 + s2 + s3 + s4)/4.0;
}

float shadowSharp15(sampler2D tex, vec3 pos) {
    float stepsize = rcp(float(shadowMapResolution));
    float n     = ditherBluenoiseStatic()*pi;
    vec2 noise  = vec2(cos(n), sin(n)) * stepsize;

    float avgDepth;

    for (uint i = 0; i < 15; ++i) {
        avgDepth   += shadowDefault(tex, pos, poisson15[i] * stepsize*2);
    }
    avgDepth /= 15.0;

    return linStep(avgDepth, 0.4, 0.6);
}

float shadowSharp30(sampler2D tex, vec3 pos) {
    float stepsize = rcp(float(shadowMapResolution));
    float n     = ditherGradNoiseTemporal()*pi;
    vec2 noise  = vec2(cos(n), sin(n)) * stepsize;

    float avgDepth;

    for (uint i = 0; i < 30; ++i) {
        avgDepth   += shadowDefault(tex, pos, poisson30[i] * stepsize*2 + noise * 0.71);
    }
    avgDepth /= 30.0;

    return linStep(avgDepth, 0.7, 0.8);
}

const vec3 waterExtinctionCoeff     = vec3(8e-2, 4e-2, 4e-2);
const vec3 waterScatterCoeff        = vec3(3e-2);
const vec3 waterAttenCoeff          = (waterExtinctionCoeff + waterScatterCoeff) * 4.0;
const vec3 waterAbsorbCoeff         = (waterExtinctionCoeff + waterScatterCoeff);

vec3 getShadowcol(sampler2D tex, vec2 coord, float distance) {
    vec4 x  = texture(tex, coord);

    if (x.a < 0.97 && x.a > 0.93) {
        //x.rgb = vec3(x.r * 2.0);
        //float caustic   = max0(1.0 - x.y + tau * x.x);
        float caustic     = (2.5 * x.z * x.z);

        vec3 absorb     = expf(-distance * waterAbsorbCoeff);
            absorb     *= mix(caustic, 1.0, expf(-distance * 0.1));
            
        x.rgb = absorb;
    } else {
        x.rgb = sqr(x.rgb);
    }

    return mix(vec3(1.0), x.rgb, x.a);
}

vec3 getShadowcol(sampler2D tex, vec2 coord) {
    vec4 x  = texture(tex, coord);
        x.rgb = sqr(x.rgb);

    if (x.a < 0.97 && x.a > 0.93) {
        x.rgb = vec3(0.1, 0.2, 1.0);
    }

    return mix(vec3(1.0), x.rgb, x.a);
}

#ifdef SOLIDPASS
vec3 getShadowmapPos(vec3 scenePos, const float bias, out float warp) {  //shadow 2d
    vec3 pos    = scenePos + vec3(bias) * lightvec;
    float a     = length(pos);
        pos     = viewMAD(shadowModelView, pos);
        pos     = projMAD(shadowProjection, pos);
        pos.z  *= 0.2;
        pos.z  -= 0.0012*(saturate(a/256.0));

        warp    = 1.0;
        pos.xy  = shadowmapWarp(pos.xy, warp);

    return pos*0.5+0.5;
}

shadowData getShadowRegular(bool diffLit, vec3 scenePos) {
    const float bias    = 0.08*(2048.0/shadowMapResolution);
    float warp          = 1.0;

    shadowData data     = shadowData(1.0, vec3(1.0), vec3(0.0));

    if (diffLit) {
        vec3 pos        = getShadowmapPos(scenePos, bias, warp);
        #ifdef shadowFilterHQ
            float s0        = getShadowSoftHQ(shadowtex0, pos);
            float s1        = getShadowSoftHQ(shadowtex1, pos);
        #else
            float s0        = getShadowSoft(shadowtex0, pos);
            float s1        = getShadowSoft(shadowtex1, pos);
        #endif

        bool translucent = distance(s0, s1)>0.1;

        data.shadow   = s1;

        if (translucent) {
            float s0depth = texture(shadowtex0, pos.xy).x;
            float dist  = distance(s0depth, pos.z) * 2.0 * 256.0 * rcp(0.2);
            data.color  = getShadowcol(shadowcolor0, pos.xy, dist);
        }
    }

    return data;
}
#else
shadowData getShadowRegular(bool diffLit) {
    shadowData data     = shadowData(1.0, vec3(1.0), vec3(0.0));

    if (diffLit) {
        vec3 pos        = shadowmapPos;
        #ifdef shadowFilterHQ
            float s0        = getShadowSoftHQ(shadowtex0, pos);
            float s1        = getShadowSoftHQ(shadowtex1, pos);
        #else
            float s0        = getShadowSoft(shadowtex0, pos);
            float s1        = getShadowSoft(shadowtex1, pos);
        #endif

        bool translucent = distance(s0, s1)>0.1;

        data.shadow   = s1;

        if (translucent) {
            float s0depth = texture(shadowtex0, pos.xy).x;
            float dist  = distance(s0depth, pos.z) * 2.0 * 256.0 * rcp(0.2);
            data.color  = getShadowcol(shadowcolor0, pos.xy, dist);
        }
        data.color      = linearToAP1Albedo(toLinear(data.color));
    }

    return data;
}
#endif