#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.github.io/license/

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

const int shadowMapResolution   = 2048;     //[512 1024 1536 2048 2560 3072 3584 4096 6144 8192 16384]
const float shadowDistance      = 128.0;

const bool shadowHardwareFiltering = true;

in vec2 coord;

flat in mat4x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex5;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D noisetex;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;

uniform int frameCounter;
uniform int isEyeInWater;
uniform int worldTime;

uniform float eyeAltitude;
uniform float far, near;
uniform float lightFlip;
uniform float sunAngle;

uniform ivec2 eyeBrightness;
uniform ivec2 eyeBrightnessSmooth;

uniform vec2 taaOffset;
uniform vec2 viewSize, pixelSize;
uniform vec2 skyCaptureResolution;

uniform vec3 cameraPosition;
uniform vec3 lightvec, lightvecView;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;


/* ------ includes ------ */
#define FUTIL_TBLEND
#define FUTIL_MAT16
#include "/lib/fUtil.glsl"

#include "/lib/frag/bluenoise.glsl"

#include "/lib/util/transforms.glsl"

#include "/lib/atmos/air/const.glsl"

#include "/lib/atmos/phase.glsl"

#include "/lib/light/warp.glsl"

#include "/lib/frag/noise.glsl"

vec3 getShadowmapPos(vec3 scenePos, const float bias) {  //shadow 2d
    vec3 pos    = scenePos + vec3(bias) * lightvec;
    float a     = length(pos);
        pos     = viewMAD(shadowModelView, pos);
        pos     = projMAD(shadowProjection, pos);
        pos.z  *= 0.2;
        pos.z  -= 0.0012*(saturate(a/256.0));

        pos.xy  = shadowmapWarp(pos.xy);

    return pos*0.5+0.5;
}
vec3 toShadowmapTexturePos(vec3 shadowPos) {
    //shadowPos.z    *= 0.2;
    shadowPos.xy    = shadowmapWarp(shadowPos.xy) * 0.5 + 0.5;

    return shadowPos;
}

const vec3 waterExtinctionCoeff     = vec3(8e-2, 4e-2, 4e-2);
const vec3 waterScatterCoeff        = vec3(3e-2);
const vec3 waterAttenCoeff          = (waterExtinctionCoeff + waterScatterCoeff) * 4.0;
const vec3 waterAbsorbCoeff         = (waterExtinctionCoeff + waterScatterCoeff);

vec3 getShadowcol(sampler2D tex, vec2 coord) {
    vec4 x  = texture(tex, coord);

    if (x.a < 0.97 && x.a > 0.93) {
        const float distance = 15.0;
        float caustic     = (2.5 * x.z * x.z);

        vec3 absorb     = expf(-distance * waterAbsorbCoeff);
            absorb     *= mix(caustic, 1.0, expf(-distance * 0.1));
            
        x.rgb = absorb;
    } else {
        x.rgb = sqr(x.rgb);
    }

    return mix(vec3(1.0), x.rgb, x.a);
}

vec3 getShadowcolUW(sampler2D tex, vec2 coord) {
    vec4 x  = texture(tex, coord);

    if (x.a < 0.97 && x.a > 0.93) {
        const float distance = 25.0;
        float caustic     = (2.5 * x.z * x.z);

        float absorb      = mix(caustic, 1.0, expf(-distance * 0.1));
            
        x.rgb = vec3(absorb);
    } else {
        x.rgb = sqr(x.rgb);
    }

    return mix(vec3(1.0), x.rgb, x.a);
}


/* ------ volumetric fog ------ */

vec2 airPhaseFunction(float cosTheta) {
    return vec2(rayleighPhase(cosTheta), mieCS(cosTheta, airMieG));
}
float airMieBackscatter(float cosTheta) {
    return mieHG(cosTheta, -airMieG * rcp(pi));
}
float fogMistPhase(float cosTheta, float density) {
    return mix(mieCS(cosTheta, pow(airMieG, 1.0 + density)), airMieBackscatter(cosTheta), 0.19);
}

uniform vec3 fogDensityCoeff;

#define fogMistAltitude 100.0

float mistDensity(vec3 rPos, float altitude) {
    float noise = value3D(rPos * 0.03) + value3D(rPos * 0.06) * 0.5;

    float mist  = expf(-max0((altitude - 64.0) * 0.01));
        mist    = sstep(noise * rcp(1.5) - (1.0 - mist) * 0.5, 0.4, 0.7);
        mist   *= sqr(1.0 - linStep(altitude, 85.0, fogMistAltitude));

    return mist * fogDensityCoeff.z;
}

vec3 fogAirDensity(vec3 rPos, float altitude) {
    rPos   += cameraPosition;

    float maxFade = sqr(1.0 - linStep(altitude, 144.0, 256.0));

    vec2 rm     = expf(-max0((altitude - 64.0) * fogFalloffScale));
    //rm.y *= 0;
    float mist  = 0.0;
    if (fogDensityCoeff.z > 1e-5) mist = mistDensity(rPos, altitude);

    return vec3(rm * fogAirScale * fogDensityCoeff.xy * maxFade, mist);
}

#define airDensity 1.0
#define fogMinSteps 8
#define fogAdaptiveSteps 12
#define fogClipDist 512.0

float fogMistLightOD(vec3 pos) {
    float stepsize = 12.0;

    const int steps = 2;

    float od = 0.0;

    for(uint i = 0; i < steps; ++i, pos += lightvec * stepsize) {

        if(pos.y > fogMistAltitude || pos.y < 0.0) continue;
        
        float density = mistDensity(pos, pos.y);
        if (density <= 0.0) continue;

            od += density * stepsize;
    }

    return od;
}

mat2x3 volumetricFog(vec3 scenePos, vec3 sceneDir, bool isSky, float vDotL, float dither, float cave) {
    float topDist    = length(sceneDir * ((256.0 - eyeAltitude) * rcp(sceneDir.y)));
    float bottomDist = length(sceneDir * ((-32.0 - eyeAltitude) * rcp(sceneDir.y)));

    float volumeHull = sceneDir.y > 0.0 ? topDist : bottomDist;

    float endDist   = isSky ? min(volumeHull, fogClipDist) : length(scenePos);
    float startDist = eyeAltitude > 256.0 ? topDist : 1.0;

    vec3 startPos   = eyeAltitude > 256.0 ? sceneDir * startDist : vec3(0.0);
        startPos   += gbufferModelViewInverse[3].xyz;
    vec3 endPos     = isSky ? sceneDir * endDist : scenePos;

    float baseStep  = length(endPos - startPos);
    float stepCoeff = saturate(baseStep * rcp(clamp(far, 256.0, 512.0)));

    uint steps      = fogMinSteps + uint(stepCoeff * fogAdaptiveSteps);

    vec3 rStep      = (endPos - startPos) / float(steps);
    vec3 rPos       = startPos + rStep * dither;
    float rLength   = length(rStep);

    vec3 shadowStartPos = viewMAD(shadowModelView, (startPos));
        shadowStartPos  = projMAD(shadowProjection, shadowStartPos);
        shadowStartPos.z *= 0.2;
    vec3 shadowEndPos   = viewMAD(shadowModelView, (endPos));
        shadowEndPos    = projMAD(shadowProjection, shadowEndPos);
        shadowEndPos.z *= 0.2;

    vec3 shadowStep = (shadowEndPos - shadowStartPos) / float(steps);
    vec3 shadowPos  = shadowStartPos + shadowStep * dither;

    mat2x3 scattering = mat2x3(0.0);
    vec3 transmittance = vec3(1.0);

    vec3 phase;
        phase.xy    = airPhaseFunction(vDotL);
        phase.z     = mix(mieCS(vDotL, airMieG * 0.8), airMieBackscatter(vDotL), 0.19);
    float phaseIso  = 0.25;

    vec3 sunlight       = (worldTime>23000 || worldTime<12900) ? lightColor[0] : lightColor[2];
        sunlight       *= pi * lightFlip;
    vec3 skylight       = lightColor[1] * cave;

    float pFade         = saturate(mieHG(vDotL, 0.65));

    for (uint i = 0; i < steps; ++i, rPos += rStep, shadowPos += shadowStep) {
        if (max3(transmittance) < 0.01) break;

        float altitude  = rPos.y + eyeAltitude;

        if (altitude > 256.0) continue;

        vec3 density    = fogAirDensity(rPos, altitude);

        //if (max3(density) < 1e-32) continue;

        vec3 stepRho    = density * rLength;
        vec3 od         = fogExtinctMat * stepRho;

        vec3 stepT      = expf(-od);
        vec3 scatterInt = saturate((stepT - 1.0) * rcp(-max(od, 1e-16)));
        vec3 visScatter = transmittance * scatterInt;

        #ifdef fogMistAdvanced
            if (density.z > 0.0) {
                float mistLightOD = fogMistLightOD(rPos + cameraPosition);

                float mistLighting  = expf(-stepRho.z) * pow(1.0 + 1.0 * 0.7 * mistLightOD, -1.0 / 0.7);
                    mistLighting    = mix(pi * (1.0 - 0.6 * expf(-density.z * 28.0)), 1.0, pFade);
                    mistLighting    = fogMistPhase(vDotL, mistLightOD);
                phase.z     = mistLighting;
            }
        #else
            if (density.z > 0.0) {
                phase.z     = fogMistPhase(vDotL, density.z) * (2.0 - pFade);
            }
        #endif

        vec3 sunScatter = fogScatterMat * (stepRho * phase) * visScatter;
        vec3 skyScatter = fogScatterMat * (stepRho * phaseIso) * visScatter;

        vec3 shadowCoord = vec3(shadowmapWarp(shadowPos.xy), shadowPos.z) * 0.5 + 0.5;

        float shadow0   = texture(shadowtex0, shadowCoord);

        float shadow    = 1.0;
        vec3 shadowCol  = vec3(1.0);
        
        if (shadow0 < 1.0) {
            shadow      = texture(shadowtex1, shadowCoord);

            if (abs(shadow - shadow0) > 0.1) {
                shadowCol   = getShadowcol(shadowcolor0, shadowCoord.xy);
            }
        }

        scattering[0]  += (sunScatter * shadowCol * transmittance) * shadow;
        scattering[1]  += skyScatter * transmittance;

        transmittance  *= stepT;
    }

    vec3 color  = scattering[0] * sunlight + scattering[1] * skylight;

    return mat2x3(color, transmittance);
}

vec3 simpleFog(vec3 scenecolor, float d, vec3 color) {
    float dist      = max(0.0, d-32.0);
    float density   = 1.0-exp(-dist*3e-3);

    return mix(scenecolor, color / pi, density);
}
vec3 waterFog(vec3 scenecolor, float d, vec3 color) {
    float dist      = max(0.0, d);
        //dist        = min(dist, 32.0);
    float density   = dist*6.5e-1;

    const vec3 absorptionCoeff  = vec3(1.0, 0.4, 0.21);
    const vec3 scatterCoeff     = vec3(0.1, 0.14, 0.3);

    vec3 scatter    = 1.0-exp(-density * scatterCoeff);
        scatter    *= max(expf(-absorptionCoeff * density), expf(-absorptionCoeff * pi));

    vec3 transmittance = exp(-density * absorptionCoeff);

    return scenecolor*transmittance + scatter * color * rcp(pi);
}

#define waterVolMinSteps 8
#define waterVolAdaptiveSteps 6
#define waterVolClipDist 48.0

mat2x3 volumetricWater(vec3 scenePos, vec3 startPos0, vec3 sceneDir, bool isSky, float vDotL, float dither, float cave) {
    float endDist   = isSky ? waterVolClipDist : length(scenePos);

    vec3 startPos   = startPos0;
        //startPos   += gbufferModelViewInverse[3].xyz;
    vec3 endPos     = isSky ? sceneDir * waterVolClipDist : scenePos;
        if (distance(startPos, endPos) > waterVolClipDist) endPos = waterVolClipDist * sceneDir;

    float baseStep  = length(endPos - startPos);
    float stepCoeff = saturate(baseStep * rcp(waterVolClipDist));

    uint steps      = waterVolMinSteps /* uint(stepCoeff * waterVolAdaptiveSteps)*/;

    vec3 rStep      = (endPos - startPos) / float(steps);
    vec3 rPos       = startPos + rStep * dither;
    float rLength   = length(rStep);

    vec3 shadowStartPos = viewMAD(shadowModelView, (startPos));
        shadowStartPos  = projMAD(shadowProjection, shadowStartPos);
        shadowStartPos.z *= 0.2;
    vec3 shadowEndPos   = viewMAD(shadowModelView, (endPos));
        shadowEndPos    = projMAD(shadowProjection, shadowEndPos);
        shadowEndPos.z *= 0.2;

    vec3 shadowStep = (shadowEndPos - shadowStartPos) / float(steps);
    vec3 shadowPos  = shadowStartPos + shadowStep * dither;

    mat2x3 scattering = mat2x3(0.0);
    vec3 transmittance = vec3(1.0);

    float phase     = mieHG(vDotL, 0.7);
    float phaseIso  = 0.25 * rcp(pi);
        phase       = mix(phase, phaseIso, 0.3);

    vec3 sunlight       = (worldTime>23000 || worldTime<12900) ? lightColor[0] : lightColor[2];
        sunlight       *= lightFlip;
    vec3 skylight       = lightColor[1] * cave;

    float pFade         = saturate(mieHG(vDotL, 0.65));

    for (uint i = 0; i < steps; ++i, rPos += rStep, shadowPos += shadowStep) {
        if (max3(transmittance) < 0.01) break;

        float altitude  = rPos.y + eyeAltitude;

        if (altitude > 256.0) continue;

        //if (distance(startPos, rPos) > waterVolClipDist) break;

        //float density   = 1.0;

        //if (max3(density) < 1e-32) continue;

        float stepRho   = rLength;
        vec3 od         = waterAttenCoeff * stepRho;

        //vec3 stepT      = expf(-od);
        //vec3 scatterInt = saturate((stepT - 1.0) * rcp(-max(od, 1e-16)));
        //vec3 visScatter = transmittance * scatterInt;

        //vec3 sunScatter = stepRho * visScatter;
        //vec3 skyScatter = stepRho * visScatter;

        vec3 shadowCoord = vec3(shadowmapWarp(shadowPos.xy), shadowPos.z) * 0.5 + 0.5;

        float shadow0   = texture(shadowtex0, shadowCoord);

        float shadow    = 1.0;
        vec3 shadowCol  = vec3(1.0);
        
        if (shadow0 < 1.0) {
            shadow      = texture(shadowtex1, shadowCoord);

            if (abs(shadow - shadow0) > 0.1) {
                shadowCol   = getShadowcol(shadowcolor0, shadowCoord.xy);
            }
        }

        scattering[0]  += (shadowCol * transmittance) * shadow;
        scattering[1]  += transmittance;

        transmittance  *= expf(-waterAttenCoeff * stepRho);
    }

    vec3 attenMul   = (1.0 - expf(-waterAttenCoeff * rLength)) * rcp(waterAttenCoeff);

    scattering[0]  *= waterScatterCoeff * attenMul * phase;
    scattering[1]  *= waterScatterCoeff * attenMul;

    vec3 color  = scattering[0] * sunlight + scattering[1] * skylight;

    return mat2x3(color, transmittance);
}

void applyFogData(inout vec3 color, in mat2x3 data) {
    color = color * data[1] + data[0];
}

/* ------ reflections ------ */

#include "/lib/light/brdf.glsl"


vec4 screenRT(vec3 viewPos, vec3 viewNormal, vec3 scenePos, vec3 sceneNormal, float dither, bool water) {
    const uint steps = 25;
    const uint refine = 5;

    vec3 viewDir        = normalize(viewPos);

    vec3 reflectDir     = reflect(viewDir, viewNormal);

    float nDotV     = saturate(1.0 - dot(viewDir, viewNormal));

    vec3 startPos   = viewPos;
    vec3 startScreenPos  = viewToScreenSpace(startPos);
    vec3 rStep      = reflectDir * mix(1.0, 0.15, nDotV);
    vec3 rPos       = startPos + rStep * dither * 0.8;
        if (!water) rPos += rStep * 0.1;

    bool hit    = false;
    bool hitBehind = false;
    float hitErrorScale = 1.0;

    vec3 hitPos = vec3(0.0);

    for (uint i = 0; i < steps; ++i, rPos += rStep) {
        vec3 screenPos  = viewToScreenSpace(rPos);

        if (saturate(screenPos.xy) != screenPos.xy) return vec4(0.0);

        float depth     = texelFetch(depthtex1, ivec2(screenPos.xy * viewSize), 0).x;
        vec3 sampleViewPos = screenToViewSpace(vec3(screenPos.xy, depth));

        float error     = distance(rPos, sampleViewPos);

        float slength   = length(rStep);

        if (error < slength && (distance(startPos, rPos) > 0.5 || water)) {
            hitBehind = screenPos.z >= depth;
            hit = true;
            hitErrorScale = (hitBehind ? slength - error : error) * rcp(slength + 1e-10);
            hitPos = rPos;
            break;
        }

        if (error > (slength * 4.0)) rStep *= 2.0;
    }

    if (hit) {
        bool refineHit  = false;

        vec3 refineStep = rStep * hitErrorScale * 0.5;
        vec3 refinePos  = rPos;
        if (hitBehind) refinePos -= rStep;

        vec3 refineHitPos = hitPos;

        for (uint j = 0; j < refine; ++j, refinePos += refineStep) {
            vec3 screenPos  = viewToScreenSpace(refinePos);

            if (saturate(screenPos.xy) != screenPos.xy) return vec4(0.0);

            float depth     = texelFetch(depthtex1, ivec2(screenPos.xy * viewSize), 0).x;

            if (screenPos.z >= depth) {
                refineHit = true;
                refineHitPos = refinePos;
                break;
            }
            refineStep *= 0.5;
        }

        vec3 stepDirection  = (refineHitPos - viewPos);

        vec3 screenPos  = viewToScreenSpace(refineHitPos);


        if (hit && landMask(texture(depthtex1, screenPos.xy).x)) {
            

            //if (dot(normalize(stepDirection), viewNormal) > 0.8) return vec4(0.0);

            

            return vec4(texture(colortex0, screenPos.xy).rgb, 1.0);
        }
    }

    return vec4(0.0);
}

vec2 unprojectSphere(vec3 dir) {
    vec2 lonlat     = vec2(atan(-dir.x, dir.z), acos(dir.y));
    return lonlat * vec2(rcp(tau), rcp(pi)) + vec2(0.5, 0.0);
}

vec3 fresnelTinted(vec3 tint, float VoH) {
    float f0    = v3avg(tint);
        f0      = min(sqrt(f0), 0.99999);
        f0      = (1.0 + f0) * rcp(1.0 - f0);

    float sinThetaI     = sqrt(saturate(1.0 - sqr(VoH)));
    float sinThetaT     = sinThetaI * rcp(f0);
    float cosThetaT     = sqrt(1.0 - sqr(sinThetaT));

    float Rs    = sqr((VoH - (f0 * cosThetaT)) * rcp(VoH + (f0 * cosThetaT)));
    float Rp    = sqr((cosThetaT - (f0 * VoH)) * rcp(cosThetaT + (f0 * VoH)));

    return saturate((Rs + Rp) * 0.5) * tint;
}

void main() {
    vec4 sceneColor   = stex(colortex0);  //that macro certainly makes it neater
    //vec4 tex1       = stex(colortex1);
    vec4 tex2       = stex(colortex2);
    vec4 tex3       = stex(colortex3);

    vec4 tex1       = stex(colortex1);
    vec2 sceneLmap  = decode2x8(tex1.z);
    vec3 sceneNormal = decodeNormal(tex1.xy);
    
    //~4fps for the whole part below
    
    int matID      = decodeMatID16(tex2.x);

    float sceneDepth0 = stex(depthtex0).x;
    vec3 viewPos0     = screenToViewSpace(vec3(coord, sceneDepth0));
    vec3 scenePos0        = viewToSceneSpace(viewPos0);

    float sceneDepth1 = stex(depthtex1).x;
    vec3 viewPos1     = screenToViewSpace(vec3(coord, sceneDepth1));
    vec3 scenePos1        = viewToSceneSpace(viewPos1);

    float vDotL     = dot(normalize(viewPos1), lightvecView);
    
    bool translucent = (sceneDepth0<sceneDepth1);

    bool water      = matID == 102;

    vec3 translucentAlbedo = sqr(decode3x8(tex2.g));

    float dist0     = distance(scenePos0, gbufferModelViewInverse[3].xyz);
    float dist1     = distance(scenePos1, gbufferModelViewInverse[3].xyz);

    float caveMult  = linStep(eyeBrightnessSmooth.y/240.0, 0.1, 0.9);

    vec3 fogcol     = (sunAngle<0.5 ? lightColor[0] : lightColor[3]) * lightFlip * mieHG(vDotL, 0.68) * sqr(caveMult);
        fogcol     *= mix(0.33, 1.0, sqrt(abs(lightvec.y)));
        fogcol     += lightColor[1] * caveMult * rcp(pi);

    float bluenoise     = ditherBluenoise();

    if (translucent && isEyeInWater==0){
        sceneColor.rgb  = blendTranslucencies(sceneColor.rgb, tex3, translucentAlbedo);
        if (water) {
            mat2x3 waterVol   = volumetricWater(scenePos1, scenePos0, normalize(scenePos0), !landMask(sceneDepth0), vDotL, bluenoise, caveMult);
            applyFogData(sceneColor.rgb, waterVol);
        }
    }

    if (water){
        vec3 viewDir    = normalize(viewPos0);
        vec3 viewNormal = normalize(mat3(gbufferModelView) * sceneNormal);
        vec3 reflectDir = reflect(viewDir, viewNormal);

        if (dot(viewDir, viewNormal) > 0.0) viewNormal = -viewNormal;

        vec4 reflection     = screenRT(viewPos0, viewNormal, scenePos0, sceneNormal, bluenoise, translucent);

        if (reflection.a < 0.5) {
            vec2 sphereCoord    = unprojectSphere(mat3(gbufferModelViewInverse) * reflectDir);
                sphereCoord    *= rcp(skyCaptureResolution);
                sphereCoord.y  += exp2(-SKY_RENDER_LOD) * 2.0;

            reflection.rgb      = texture(colortex5, sphereCoord).rgb;
            reflection.a        = 1.0;
        }

        materialProperties mat  = materialProperties(0.0001, 0.02, false, mat2x3(1.0));

        vec3 shadowPos  = getShadowmapPos(scenePos0, 0.08);
        float shadow    = texture(shadowtex1, shadowPos);
        
        vec3 fresnel   = BRDFfresnel(-viewDir, viewNormal, mat);
        //fresnel     = fresnelTinted(translucentAlbedo, dot(-viewDir, viewNormal));
        
        sceneColor.rgb      = mix(sceneColor.rgb, reflection.rgb, vec3(reflection.a) * fresnel);
        if (shadow > 0.01) {
            vec3 directCol  = (sunAngle<0.5 ? lightColor[0] : lightColor[2] * 0.1) * lightFlip;
            sceneColor.rgb += BRDF(-viewDir, lightvecView, viewNormal, mat) * directCol;
        }

        //sceneColor.rgb = reflection.rgb;
    }

    if (isEyeInWater==1) {
        sceneColor.rgb  = blendTranslucencies(sceneColor.rgb, tex3, translucentAlbedo);
        
        mat2x3 waterVol   = volumetricWater(scenePos0, gbufferModelViewInverse[3].xyz, normalize(scenePos0), !landMask(sceneDepth0), vDotL, bluenoise, caveMult);
        applyFogData(sceneColor.rgb, waterVol);
    }

    #ifdef fogVolumeEnabled
    if (isEyeInWater==0) {
        mat2x3 fogData  = volumetricFog(scenePos0, normalize(scenePos0), !landMask(sceneDepth0), vDotL, bluenoise, caveMult);

        applyFogData(sceneColor.rgb, fogData);
    }
    #endif

    if (matID == 3) {
        if (landMask(sceneDepth0)) sceneColor.rgb = sceneColor.rgb*0.6 + vec3(0.5*v3avg(sceneColor.rgb));
        else sceneColor.rgb = sceneColor.rgb*0.7 + vec3(0.8*v3avg(sceneColor.rgb));
    }

    #ifdef compNAN
    if (isnan3(sceneColor.rgb) || isinf3(sceneColor.rgb)) sceneColor.g = 100.0;
    #endif

    //sceneColor.rgb = sqr(sceneNormal * 0.5 + 0.5);

    /*DRAWBUFFERS:0*/
    gl_FragData[0]  = makeDrawbuffer(sceneColor);
}