#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.github.io/license/

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/colorspace.glsl"

const int shadowMapResolution   = 2048;     //[512 1024 1536 2048 2560 3072 3584 4096 6144 8192 16384]
const float shadowDistance      = 128.0;

const bool shadowHardwareFiltering = true;

in vec2 coord;

flat in mat3x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;

uniform sampler2D noisetex;

uniform sampler2D depthtex1;

uniform sampler2DShadow shadowtex0;
uniform sampler2DShadow shadowtex1;
uniform sampler2D shadowcolor0;

uniform int frameCounter;

uniform float aspectRatio;
uniform float far, near;
uniform float sunAngle;
uniform float viewHeight, viewWidth;

uniform vec2 taaOffset;
uniform vec2 viewSize;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;


/* ------ includes ------ */
#define FUTIL_LINDEPTH
#define FUTIL_LIGHTMAP
#define FUTIL_ROT2
#define FUTIL_MAT16
#include "/lib/fUtil.glsl"

#include "/lib/util/transforms.glsl"

#include "/lib/frag/bluenoise.glsl"

vec3 lightvec     = normalize(vec3(0.0, 1.0, 0.0));
vec3 lightvecView = vec3(0.0);

#include "/lib/light/diffuse.glsl"

#include "/lib/light/warp.glsl"

#define SOLIDPASS

#include "/lib/light/shadow.glsl"
#include "/lib/atmos/phase.glsl"
#include "/lib/light/subsurface.glsl"

vec4 textureBilateral(sampler2D tex, sampler2D depth, const int lod, float fdepth) {
    vec4 data   = vec4(0.0);
    float sum   = 0.0;
    ivec2 posD  = ivec2(coord*viewSize);
    ivec2 posT  = ivec2(coord*viewSize*rcp(float(lod)));
    vec3 zmult  = vec3((far*near)*2.0, far+near, far-near);
        fdepth  = depthLinear(fdepth);
    
    for (int i = -1; i<2; i++) {
        for (int j = -1; j<2; j++) {
            ivec2 tcDepth = posD + ivec2(i, j)*lod;
            float dsample = depthLinear(texelFetch(depth, tcDepth, 0).x);
            float w     = abs(dsample-fdepth)*zmult.x<1.0 ? 1.0 : 1e-5;
            ivec2 ct    = posT + ivec2(i, j);
            data       += texelFetch(tex, ct, 0)*w;
            sum        += w;
        }
    }
    data *= rcp(sum);

    return data;
}

void main() {
    vec4 sceneColor   = stex(colortex0);
    float sceneDepth = stex(depthtex1).x;

    if (landMask(sceneDepth)) {
        lightvec.yz     = rotatePos(lightvec.yz, radians(sunPathRotation));
        lightvec        = normalize(lightvec);

        lightvecView    = mat3(gbufferModelView) * lightvec;

        vec4 tex1       = stex(colortex1);
        vec2 sceneLmap  = decode2x8(tex1.z);
        vec3 sceneNormal = decodeNormal(tex1.xy);

        vec4 tex2       = stex(colortex2);
        int matID       = decodeMatID16(tex2.x);

        #if DEBUG_VIEW==1
        sceneColor.rgb    = vec3(0.25);
        #endif

        vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth));
        vec3 scenePos   = viewToSceneSpace(viewPos);

        #ifdef lightmapSmoothingEnabled
        vec3 tex3       = stex(colortex3).xyz;
            sceneLmap   = tex3.xy;
            sceneColor.a = tex3.z;
        #endif

        float ao        = sqr(sceneColor.a)*0.66+0.34;
        
        #ifdef ambientOcclusionEnabled
            vec4 upscale3 = textureBilateral(colortex3, depthtex1, 2, sceneDepth);
                ao     *= upscale3.w;
        #endif
        
        /* ------ lighting calculation ------ */
        //~6 fps
        
        vec2 lmap       = sceneLmap;
        //lmap.y          = cube(lmap.y);

        float diff      = diffuseLambert(sceneNormal);
        #ifndef subsurfaceScatteringEnabled
            if (matID == 2) diff = diff*0.3+0.7;
            if (matID == 4) diff = sqrt(diff);
        #endif

        shadowData shadow = shadowData(1.0, vec3(1.0), vec3(0.0));

        #ifdef subsurfaceScatteringEnabled
            if (matID == 2 || matID == 4) shadow = getShadowSubsurface(diff>0.0, scenePos, normalize(viewPos), sceneColor.rgb);
            else shadow = getShadowRegular(diff>0.0, scenePos);    //~3fps, almost for free
        #else
            shadow = getShadowRegular(diff>0.0, scenePos);    //~3fps, almost for free
        #endif

        float diffLit       = min(diff, shadow.shadow);
        vec3 directCol      = lightColor[0];
        vec3 directLight    = diffLit * shadow.color * directCol;

        #ifdef subsurfaceScatteringEnabled
            directLight    += shadow.subsurfaceScatter * directCol;
        #endif

        vec3 indirectLight  = lightColor[1];
            indirectLight  += vec3(0.5, 0.7, 1.0) * (0.01 * minLightLuma);
            indirectLight  *= ao;

        vec3 result     = directLight + indirectLight;
            result     += getBlocklightMap(lightColor[2], lmap.x)*ao;

        if (matID == 5) {
            vec3 emitterCol     = mix(lightColor[2], vec3(v3avg(lightColor[2])), emitterRecoloring);
                emitterCol     *= cubeSmooth(saturate(v3avg(sceneColor.rgb) * 1.5));
            result += emitterCol * pi;
        }

        sceneColor.rgb   *= result;
        //sceneColor.rgb    = shadow.subsurfaceScatter;

        #if DEBUG_VIEW==2
        sceneColor.rgb    = ao * lightColor[0];
        #endif
    }

    /*DRAWBUFFERS:03*/
    gl_FragData[0]  = makeDrawbuffer(sceneColor);
    gl_FragData[1]  = vec4(0.0);
}